//#include "xbasic_types.h"
//#include "xio.h"
//#include "ether.h"

#ifndef BYTE
#define BYTE unsigned char
#endif
#ifndef WORD
#define WORD unsigned short
#endif
#ifndef LWORD
#define LWORD unsigned long
#endif

/* Ensure network structures aren't padded (DJGPP and VC) */
#pragma pack(1)

#define MACLEN      6           /* Ethernet (MAC) address length */
#define BCASTADDR 0xff,0xff,0xff,0xff,0xff,0xff /* Broadcast address */

/* Ethernet Frame sizes */
#define MAXFRAME    1514        /* Maximum frame size (inc hdr)*/
#define MINFRAME    64          /* Minimum frame size (inc hdr) */
#define ETHERMTU    80   //(MAXFRAME-sizeof(ETHERHDR))

/* Ethernet (DIX) header */
typedef struct {
    BYTE dest[MACLEN];          /* Destination MAC address */
    BYTE srce[MACLEN];          /* Source MAC address */
    WORD ptype;                 /* Protocol type or length */
} ETHERHDR;

/* Ethernet frame; data size is frame size minus header & CRC */
typedef struct {
    ETHERHDR h;                 /* Header */
    BYTE data[ETHERMTU];        /* Data */
} ETHERFRAME;

/* NE2000-compatible driver prototypes */
int init_etherne();
void poll_etherne();
WORD get_etherne(void *pkt);
WORD put_etherne(void *pkt, WORD len);

#define WORDMODE 1 // Set to zero if using 8-bit bus

// NE2000 definitions
#define NIC_BASE (0x00A00400) // Base I/O address of the NIC card
#define DATAPORT (0x10*2)
#define NE_RESET (0x1f*2)

// NIC page0 register offsets
#define CMDR    (0x00*2) // command register for read & write
#define PSTART  (0x01*2) // page start register for write
#define PSTOP   (0x02*2) // page stop register for write
#define BNRY    (0x03*2) // boundary reg for rd and wr
#define TPSR    (0x04*2) // tx start page start reg for wr
#define TBCR0   (0x05*2) // tx byte count 0 reg for wr
#define TBCR1   (0x06*2) // tx byte count 1 reg for wr
#define ISR     (0x07*2) // interrupt status reg for rd and wr
#define RSAR0   (0x08*2) // low byte of remote start addr
#define RSAR1   (0x09*2) // hi byte of remote start addr
#define RBCR0   (0x0A*2) // remote byte count reg 0 for wr
#define RBCR1   (0x0B*2) // remote byte count reg 1 for wr
#define RCR     (0x0C*2) // rx configuration reg for wr
#define TCR     (0x0D*2) // tx configuration reg for wr
#define DCR     (0x0E*2) // data configuration reg for wr
#define IMR     (0x0F*2) // interrupt mask reg for wr

// NIC page 1 register offsets
#define PAR0    (0x01*2) // physical addr reg 0 for rd and wr
#define CURR    (0x07*2) // current page reg for rd and wr
#define MAR0    (0x08*2) // multicast addr reg 0 for rd and WR

// Buffer Length and Field Definition Info
#define TXSTART  0x40              // Tx buffer start page
#define TXPAGES  6                 // Pages for Tx buffer
#define RXSTART  (TXSTART+TXPAGES) // Rx buffer start page
#if WORDMODE
#define RXSTOP   0x7e              // Rx buffer end page for word mode
#define DCRVAL   0x49              // DCR values for word mode
#else
#define RXSTOP   0x5f              // Ditto for byte mode
#define DCRVAL   0x48
#endif
#define STARHACK 0                  /// Set non-zero to enable Starlan length hack

typedef struct        // Net driver configuration data
{
  BYTE myeth[MACLEN]; // MAC (Ethernet) addr
  WORD next_pkt;      // Next (current) Rx page
} CONFIGNE;

static CONFIGNE nic; // driver condifuration data
int promisc=0;       // Flag to enable promiscuous mode

typedef struct {     // NIC hardware packet header
  BYTE stat;         // Error status
  BYTE next;         // Pointer to next block
  WORD len;          // Length of this frame incl. CRC
} NICHDR;
NICHDR nichdr;

// Private prototypes
void resetnic(CONFIGNE *cp, char cold);
void getnic(WORD addr, BYTE data[], WORD len);
void putnic(WORD addr, BYTE data[], WORD len);
BYTE nicwrap(int page);
BYTE innic(int reg);
void outnic(int reg, int b);
void delay(int mult);

// Initialise card given driver type and base addr.
// Return driver type, 0 if error
int init_etherne()
{
  int ok=0;

  outnic(NE_RESET, innic(NE_RESET));  // Do reset
  delay(2);
  if ((innic(ISR) & 0x80) == 0)       // Report if failed
  {
    print("  Ethernet card failed to reset!\r\n");
  }
  else
  {
    resetnic(&nic, 1);                // Reset Ethernet card, get my addr
    ok = 1;
  }
  return(ok);
}
 
// Poll network interface to keep it alive; send & receive frames
void poll_etherne(){
  WORD len;
  static BYTE ebuff[MAXFRAME];

  outnic(ISR, 0x01);                  // Clear interrupt flag
  /*
  // Receive
  while ((len=get_etherne(ebuff))>0)
  {                                   // Store frames in buff
    receive_upcall(ebuff, len);
  }
  // Transmit
  while (!(innic(CMDR)&0x04) &&       // While NIC ready & frame avail
	 (len=transmit_upcall(ebuff, MAXFRAME))>0)
  {                                   // ..transmit frame
    put_etherne(ebuff, len);
  }
  */
}

// Get packet into buffer, return length (excl CRC), or 0 if none available
WORD get_etherne(void *pkt){
  WORD len=0, curr;
  BYTE bnry;
  CONFIGNE *cp;
  
  cp = &nic;
  if (innic(ISR) & 0x10)              // If Rx overrun..
  {
    print("  NIC Rx overrun\r\n");
    resetnic(cp, 0);                // ..reset controller (drastic!)
  }
  outnic(CMDR, 0x60);                 // DMA abort, page 1
  curr = innic(CURR);                 // Get current page
  outnic(CMDR, 0x20);                 // DMA abort, page 0
  if (curr != cp->next_pkt)           // If Rx packet..
  {
    nichdr.len = 0xee;
    getnic((WORD)(cp->next_pkt<<8), (BYTE *)&nichdr, sizeof(nichdr));
    
    len = nichdr.len;               // Take length from stored header
    if ((nichdr.stat&1) && len>=MINFRAME && len<=MAXFRAME)
    {                               // If hdr is OK, get packet
      if (pkt)
	getnic((WORD)((cp->next_pkt<<8)+sizeof(nichdr)), pkt, len);
    }
    else                            // If not, no packet data
    {
      print("  NIC packet error\r\n");
    }                               // Update next packet ptr
    if (nichdr.next>=RXSTART && nichdr.next<RXSTOP)
      cp->next_pkt = nichdr.next;
    else                            // If invalid, use prev+1
    {
      print("  NIC pointer error\r\n");
      cp->next_pkt = nicwrap(cp->next_pkt + 1);
    }                               // Update boundary register
    bnry = nicwrap(cp->next_pkt - 1);
    outnic(BNRY, bnry);
  }
  return(len);                        // Return length excl. CRC
}

// Send Ethernet packet given len, return 0 if NIC is busy
WORD put_etherne(void *pkt, WORD len)
{
  if (innic(CMDR) & 0x4)      // If still Txing, return 0
    len = 0;
  else if (pkt)
  {                                   // If last Tx is complete..
    outnic(ISR, 0x0a);              // Clear interrupt flags
    outnic(TBCR0, len & 0xff);      // Set Tx length regs
    outnic(TBCR1, len >> 8);
    putnic(TXSTART<<8, pkt, len);
    outnic(CMDR, 0x24);             // Transmit the packet
  }
  return(len);
}

// Reset the Ethernet card, if 'cold' start, get my 6-byte address
void resetnic(CONFIGNE *cp, char cold){
  int i;
  BYTE temp[MACLEN*2];
  
  outnic(CMDR, 0x21);                 // Stop, DMA abort, page 0
  delay(2);                           // ..wait to take effect
  outnic(DCR, DCRVAL);
  outnic(RBCR0, 0);                   // Clear remote byte count
  outnic(RBCR1, 0);
  outnic(RCR, 0x20);                  // Rx monitor mode
  outnic(TCR, 0x02);                  // Tx internal loopback
  outnic(TPSR, TXSTART);              // Set Tx start page
  outnic(PSTART, RXSTART);            // Set Rx start, stop, boundary
  outnic(PSTOP, RXSTOP);
  outnic(BNRY, (BYTE)(RXSTOP-1));
  outnic(ISR, 0xff);                  // Clear interrupt flags
  outnic(IMR, 0);                     // Mask all interrupts
  if (cold)
  {
    outnic(CMDR, 0x22);             // Start NIC, DMA abort
    getnic(0, temp, 12);            // Get 6-byte addr
    for (i=0; i<MACLEN; i++)        // Convert addr words to bytes
      cp->myeth[i] = temp[WORDMODE ? i+i : i];
  }
  outnic(CMDR, 0x61);                 // Stop, DMA abort, page 1
  delay(2);
  for (i=0; i<6; i++)                 // Set Phys addr
    outnic(PAR0+i, cp->myeth[i]);
  for (i=0; i<8; i++)                 // Multicast accept-all
    outnic(MAR0+i, 0xff);
  outnic(CURR, RXSTART+1);            // Set current Rx page
  cp->next_pkt = RXSTART + 1;
  outnic(CMDR, 0x20);                 // DMA abort, page 0
  outnic(RCR, promisc ? 0x14 : 0x04); // Allow broadcasts, maybe all pkts
  outnic(TCR, 0);                     // Normal Tx operation
  outnic(ISR, 0xff);                  // Clear interrupt flags
  outnic(CMDR, 0x22);                 // Start NIC
}

// Get a packet from a given address in the NIC's RAM
void getnic(WORD addr, BYTE data[], WORD len){
    register int count;
    register WORD *dataw, dataport;

    count = WORDMODE ? len>>1 : len;    // Halve byte count if word I/P
    outnic(ISR, 0x40);                  // Clear remote DMA interrupt flag
    outnic(RBCR0, len&0xff);            // Byte count
    outnic(RBCR1, len>>8);
    outnic(RSAR0, addr&0xff);           // Data addr
    outnic(RSAR1, addr>>8);
    outnic(CMDR, 0x0a);                 // Start, DMA remote read

#if WORDMODE
    dataw = (WORD *)data;               // Use pointer for speed
    while(count--)                      // Get words
        *dataw++ = innic(DATAPORT);
    if (len & 1)                        // If odd length, do last byte
        *(BYTE *)dataw = innic(DATAPORT);
#else
    while(count--)                      // Get bytes
        *data++ = innic(DATAPORT);
#endif

}

// Put a packet into a given address in the NIC's RAM
void putnic(WORD addr, BYTE data[], WORD len){
  register int count;
  register WORD *dataw, dataport;
  
  len += len & 1;                     // Round length up to an even value
  count = WORDMODE ? len>>1 : len;    // Halve byte count if word O/P
  outnic(ISR, 0x40);                  // Clear remote DMA interrupt flag
  outnic(RBCR0, len&0xff);            // Byte count
  outnic(RBCR1, len>>8);
  outnic(RSAR0, addr&0xff);           // Data addr
  outnic(RSAR1, addr>>8);
  outnic(CMDR, 0x12);                 // Start, DMA remote write
#if WORDMODE                            // Word transfer?
  dataw = (WORD *)data;
  while(count--)
    outnic(DATAPORT, *dataw++);  // O/P words
#else
  while(count--)                  // O/P bytes
    outnic(DATAPORT, *data++);
#endif
  count = 10000;                      // Done: must ensure DMA complete
  while(count && (innic(ISR)&0x40)==0)
    count--;
}

// Wrap an NIC Rx page number
BYTE nicwrap(int page){
  if (page >= RXSTOP)
    page += RXSTART - RXSTOP;
  else if (page < RXSTART)
    page += RXSTOP - RXSTART;
  return(page);
}

// Input a byte from a NIC register
BYTE innic(int reg){
  //return(XIo_In16(NIC_BASE+reg));
  return 0;
}

// Output a byte to a NIC register
void outnic(int reg, int b)
{
  //XIo_Out16((WORD)(NIC_BASE+reg), b);
}

void delay(int mult){
  int i;
  int delay = 1000000*mult;
  for(i=0; i<delay; i++);
}
